// Copyright David Abrahams, Daniel Wallin 2003. Use, modification and 
// distribution is subject to the Boost Software License, Version 1.0. 
// (See accompanying file LICENSE_1_0.txt or copy at 
// http://www.boost.org/LICENSE_1_0.txt)
//
//
// Modified by Roland Pihlakas (2007) 
//


#ifndef BOOST_PARAMETER_MACROS_050412_HPP
#define BOOST_PARAMETER_MACROS_050412_HPP

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/repetition/repeat_from_to.hpp>
#include <boost/preprocessor/arithmetic/inc.hpp>
#include <boost/preprocessor/logical/bool.hpp>
#include <boost/preprocessor/punctuation/comma_if.hpp>
#include <boost/preprocessor/control/expr_if.hpp>
#include <boost/preprocessor/repetition/enum_params.hpp>
#include <boost/preprocessor/repetition/enum_binary_params.hpp>
#include <boost/preprocessor/cat.hpp>

#define BOOST_PARAMETER_FUN_TEMPLATE_HEAD1(n) \
    template<BOOST_PP_ENUM_PARAMS(n, class T)>

#define BOOST_PARAMETER_FUN_TEMPLATE_HEAD0(n)

#ifndef BOOST_NO_SFINAE

# define BOOST_PARAMETER_MATCH_TYPE(n, param)           \
            BOOST_PP_EXPR_IF(n, typename) param::match  \
            <                                           \
                BOOST_PP_ENUM_PARAMS(n, T)              \
            >::type 

#else

# define BOOST_PARAMETER_MATCH_TYPE(n, param) param

#endif

#define BOOST_PARAMETER_FUN_DECL(z, n, params)                                      \
                                                                                    \
    BOOST_PP_CAT(BOOST_PARAMETER_FUN_TEMPLATE_HEAD, BOOST_PP_BOOL(n))(n)            \
                                                                                    \
    BOOST_PP_TUPLE_ELEM(3, 0, params)                                               \
        BOOST_PP_TUPLE_ELEM(3, 1, params)(                                          \
            BOOST_PP_ENUM_BINARY_PARAMS(n, T, const& p)                             \
            BOOST_PP_COMMA_IF(n)                                                    \
            BOOST_PARAMETER_MATCH_TYPE(n,BOOST_PP_TUPLE_ELEM(3, 2, params))         \
            kw = BOOST_PP_TUPLE_ELEM(3, 2, params)()                                \
        )                                                                           \
    {                                                                               \
        return BOOST_PP_CAT(BOOST_PP_TUPLE_ELEM(3, 1, params), _with_named_params)( \
            kw(BOOST_PP_ENUM_PARAMS(n, p))                                          \
        );                                                                          \
    }

// Generates:
//
// template<class Params>
// ret name ## _with_named_params(Params const&);
//
// template<class T0>
// ret name(T0 const& p0, typename parameters::match<T0>::type kw = parameters())
// {
//     return name ## _with_named_params(kw(p0));
// }
//
// template<class T0, ..., class TN>
// ret name(T0 const& p0, ..., TN const& PN
//    , typename parameters::match<T0, ..., TN>::type kw = parameters())
// {
//     return name ## _with_named_params(kw(p0, ..., pN));
// }
//
// template<class Params>
// ret name ## _with_named_params(Params const&)
//
// lo and hi determines the min and max arity of the generated functions.

#define BOOST_PARAMETER_FUN(ret, name, lo, hi, parameters)                          \
                                                                                    \
    template<class Params>                                                          \
    ret BOOST_PP_CAT(name, _with_named_params)(Params const&);                      \
                                                                                    \
    BOOST_PP_REPEAT_FROM_TO(                                                        \
        lo, BOOST_PP_INC(hi), BOOST_PARAMETER_FUN_DECL, (ret, name, parameters))    \
                                                                                    \
    template<class Params>                                                          \
    ret BOOST_PP_CAT(name, _with_named_params)(Params const& p)

#define BOOST_PARAMETER_MEMFUN(ret, name, lo, hi, parameters)                       \
                                                                                    \
    BOOST_PP_REPEAT_FROM_TO(                                                        \
        lo, BOOST_PP_INC(hi), BOOST_PARAMETER_FUN_DECL, (ret, name, parameters))    \
                                                                                    \
    template<class Params>                                                          \
    ret BOOST_PP_CAT(name, _with_named_params)(Params const& p)

#endif // BOOST_PARAMETER_MACROS_050412_HPP





//roland start


///###################################################################################
///
///###################################################################################


#define PARAMETER_TYPE(name)							typename boost::parameter::binding<Params, tag::name>::type	
#define PARAMETER_TYPE_D(name, def_T)					typename boost::parameter::binding<Params, tag::name, def_T >::type

#define PARAMETER_TYPEF(name, typeout)					typedef typename boost::parameter::binding<Params, tag::name>::type				typeout;
#define PARAMETER_TYPEF_D(name, typeout, def_T)			typedef typename boost::parameter::binding<Params, tag::name, def_T >::type	typeout;

//#define PARAMETER_TYPE_NT(name)							boost::parameter::binding<Params, tag::name>::type	
//#define PARAMETER_TYPE_D_NT(name, def_T)				boost::parameter::binding<Params, tag::name, def_T >::type



#define PARAMETER_TYPE_NR(name)							typename boost::parameter::binding_noref<Params, tag::name>::type
#define PARAMETER_TYPE_D_NR(name, def_T)				typename boost::parameter::binding_noref<Params, tag::name, def_T >::type

#define PARAMETER_TYPEF_NR(name, typeout)				typedef typename boost::parameter::binding_noref<Params, tag::name>::type			typeout;
#define PARAMETER_TYPEF_D_NR(name, typeout, def_T)		typedef typename boost::parameter::binding_noref<Params, tag::name, def_T >::type	typeout;

#define PARAMETER_TYPE_NT(name)							boost::parameter::binding_noref<Params, tag::name>::type
#define PARAMETER_TYPE_D_NT(name, def_T)				boost::parameter::binding_noref<Params, tag::name, def_T >::type





#define P_ISSET(key)									Params::isset<tag::key>::rval

#define IF_P_ISSET(key)									IFC(P_ISSET(key))						///use when you want to use else statement
#define IFN_P_ISSET(key)								IFC(!P_ISSET(key))


#define IF_P_ISSET_S(key)								IFC_S(P_ISSET(key))
#define IFN_P_ISSET_S(key)								IFNC_S(P_ISSET(key))

#define IF_P_ISSET_S2(key1, key2)						IFC_S(P_ISSET(key1) || P_ISSET(key2))
#define IFN_P_ISSET_S2(key1, key2)						IFNC_S(P_ISSET(key1) || P_ISSET(key2))



#define UNSET_P(p, key)									p.unset(key)  //p.unset<tag::key>()				//Params::unset<tag::key>(p)



/*
#define P_LAZYV(key, lazyfunc)							(P_ISSET(key) ? p[key] : lazyfunc)
//#define P_LAZY(outname, key, lazyfunc)					BOOST_AUTO(outname, P_LAZYV(key, lazyfunc));
//#define P_LAZY(outname, key, lazyfunc)					BOOST_TYPEOF(P_LAZYV(key, lazyfunc)) outname = P_LAZYV(key, lazyfunc);
#define P_LAZY(outname, key, lazyfunc)					BOOST_LVALUE_TYPEOF(P_LAZYV(key, lazyfunc)) outname = P_LAZYV(key, lazyfunc);
*/

#define P_LAZY(outname, key, lazyfunc)										\
																			\
	IF_P_ISSET_S(key)														\
	{																		\
		P_GET(outname, key);												\
	}																		\
	IFN_P_ISSET_S(key)														\
	{																		\
		BOOST_LVALUE_TYPEOF(lazyfunc) outname = lazyfunc;					\
	}																		\


#define P_GET(outname, key)								PARAMETER_TYPE(key) outname = p[key];
#define P_GET_											P_GET
#define P_GETD(outname, key, def)						PARAMETER_TYPE_D(key, BOOST_TYPEOF(def)) outname = p[key | def];
#define P_GETD2(outname, key, def_T, def)				PARAMETER_TYPE_D(key, def_T) outname = p[key | def];


#define BOOST_AUTO_LVALUE(var, expr)					BOOST_LVALUE_TYPEOF(expr) var = expr


///###################################################################################
///
///###################################################################################


#define NO_PARAMS				boost::parameter::aux::empty_arg_list()



#define P_MEMFUN2(returntype, name/*, ...*/)	template <typename Params> returntype name(Params const &p)		
//#define P_MEMFUN(...)						P_MEMFUN2(__VA_ARGS__,)					   ///allows macro calls with two or one merged macro arguments


///global functions require declaring following template function
#define P_FUN(returntype, name)										\
																	\
	template <typename Params> returntype name(Params const &p);	\
																	\
	returntype name()												\
	{																\
		return name(NO_PARAMS);										\
	};																\
																	\
	template <typename Params> returntype name(Params const &p)		//TODO: for some strange reason MSVC compiler does not break __VA_ARGS__ into pieces even when it contains commas




#define P_MEMFUN_O_DECL(returntype, name)							\
																	\
	returntype name();												\
	P_MEMFUN_CI(returntype, name);								    \


///member functions forbid declaring following template function
#define P_MEMFUN_O(returntype, name)								\
																	\
	returntype name()												\
	{																\
		return name(NO_PARAMS);										\
	};																\
																	\
	P_MEMFUN_I(returntype, name)
	
	
	
	
#define P_MEMFUN_DECL									P_MEMFUN_I(returntype, name);
	
	
#define P_MEMFUN(returntype, name)									\
																	\
	returntype name()												\
	{																\
		BOOST_STATIC_ASSERT(false);	/* prevent instantiation */		\
	};																\
																	\
	P_MEMFUN_I(returntype, name)									\
	
	

#define P_MEMFUN_I(returntype, name)					template <typename Params> returntype name(Params const &p)				//TODO: for some strange reason MSVC compiler does not break __VA_ARGS__ into pieces even when it contains commas



#define P_MEMFUN_CO_DECL(returntype, name)							\
																	\
	returntype name() const;										\
	P_MEMFUN_CI(returntype, name);								    \


#define P_MEMFUN_CO(returntype, name)								\
																	\
	returntype name() const											\
	{																\
		return name(NO_PARAMS);										\
	};																\
																	\
	P_MEMFUN_CI(returntype, name)									//TODO: for some strange reason MSVC compiler does not break __VA_ARGS__ into pieces even when it contains commas






#define P_MEMFUN_C_DECL(returntype, name)					P_MEMFUN_CI(returntype, name);


#define P_MEMFUN_C(returntype, name)								\
																	\
	returntype name() const											\
	{																\
		BOOST_STATIC_ASSERT(false);	/* prevent instantiation */		\
	};																\
																	\
	P_MEMFUN_CI(returntype, name)									\


#define P_MEMFUN_CI(returntype, name)					template <typename Params> returntype name(Params const &p) const				//TODO: for some strange reason MSVC compiler does not break __VA_ARGS__ into pieces even when it contains commas



//#define SP(x)			(x, dummy_P = 0)						///macro for single parameter. We need to put the comma operator in the call in order for boost::parameter to work


///###################################################################################
///
///###################################################################################


//roland end
